/*
Copyright (c) 2020 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <limits.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdlib.h>
#include "cstr.h"

#if defined(_WIN32) || defined(__CYGWIN__) || (defined(_POSIX_C_SOURCE) && _POSIX_C_SOURCE >= 1) || defined(_XOPEN_SOURCE) || defined(_POSIX_SOURCE)
#  include <sys/stat.h>
#  include <sys/types.h>
#endif

#if defined(_MSC_VER) && _MSC_VER >= 1914
#  pragma warning(push)
#  pragma warning(disable : 5045) // Information that MS compiler will insert Spectre mitigation if /Qspectre switch specified
#endif

// Allocated pointers point to mutable memory by design, regardless of any const declaration in the code. A simple cast to `void*` may still cause a warning though.
static inline void *discard_const(const void *ptr)
{
  const union
  {
    const void *ptr2const;
    void *ptr2mutable;
  } discardconst = { ptr };

  return discardconst.ptr2mutable;
}

// Try to determine the size of a file.
// Return true if the function succeeded. Return false if the function failed (e.g. because stdin was passed).
static bool file_size(FILE *stream, size_t *psize)
{
#if defined(_WIN32) || defined(__CYGWIN__)
  struct _stat64 data = { 0 };
  const int fd = _fileno(stream);
  if (fd != _fileno(stdin) && _fstat64(fd, &data) == 0 && data.st_size >= 0 && (size_t)data.st_size == (uint64_t)data.st_size)
  {
    *psize = (size_t)data.st_size;
    return true;
  }
#elif (defined(_POSIX_C_SOURCE) && _POSIX_C_SOURCE >= 1) || defined(_XOPEN_SOURCE) || defined(_POSIX_SOURCE)
  struct stat data = { 0 };
  const int fd = fileno(stream);
  if (fd != fileno(stdin) && fstat(fd, &data) == 0 && data.st_size >= 0)
  {
    *psize = (size_t)data.st_size;
    return true;
  }
#else
  long size = 0L;
  if (stream != stdin && fseek(stream, 0L, SEEK_END) == 0 && (size = ftell(stream)) >= 0L)
  {
    *psize = (size_t)size;
    return true;
  }
#endif

  return false;
}

// validate UTF-8 and advance the pointer to the next codepoint.
static bool u8_next_codepoint(const unsigned char **ptr)
{
  const unsigned char *it = *ptr;
  if (*it > 0x7FU)
  {
    if (*it >= 0xC2U && *it <= 0xDFU &&
        it[1] >= 0x80U && it[1] <= 0xBFU)
      ++it;
    else if (((*it == 0xE0U &&
               it[1] >= 0xA0U && it[1] <= 0xBFU) ||
              (*it >= 0xE1U && *it <= 0xECU &&
               it[1] >= 0x80U && it[1] <= 0xBFU) ||
              (*it == 0xEDU &&
               it[1] >= 0x80U && it[1] <= 0x9FU) ||
              (*it >= 0xEEU && *it <= 0xEFU &&
               it[1] >= 0x80U && it[1] <= 0xBFU)) &&
             it[2] >= 0x80U && it[2] <= 0xBFU)
      it += 2;
    else if (((*it == 0xF0U &&
               it[1] >= 0x90U && it[1] <= 0xBFU) ||
              (*it >= 0xF1U && *it <= 0xF3U &&
               it[1] >= 0x80U && it[1] <= 0xBFU) ||
              (*it == 0xF4U &&
               it[1] >= 0x80U && it[1] <= 0x8FU)) &&
             it[2] >= 0x80U && it[2] <= 0xBFU &&
             it[3] >= 0x80U && it[3] <= 0xBFU)
      it += 3;
    else
    {
      *ptr = it + 1;
      return false;
    }
  }

  *ptr = it + 1;
  return true;
}

// Execution of CSDIE causes the program to exit abnormally.
#if defined(_MSC_VER)
#  define CSDIE __debugbreak()
#elif defined(__clang__) || defined(__GNUC__)
#  define CSDIE __builtin_trap()
#else
#  define CSDIE ((void)(*(volatile char *)0 = 0))
#endif

// Check if the sum fits into a limited range, die if not.
static inline size_t cstr_add_or_die(const size_t ubound, const size_t a, const size_t b)
{
  if (a > ubound || b > ubound - a)
    CSDIE;

  return a + b;
}

//////////////////////////////////////////////////////////
// vvv  based on char  vvv

// A rather abstract structure. It's never instantiated but makes it easier to access the information in the allocated memory space.
struct cstr_base
{
  size_t capacity;
  size_t size;
  char data[sizeof(size_t)]; // Placeholder for the actual cstr. The array declaration here requires some decay hacks for analysis tools to calm down.
};

// Actually only these pointers to cstr_base are used.
typedef struct cstr_base *pcsbase;
typedef const struct cstr_base *pccsbase;

// Distance in bytes of the data member from the beginning of cstr_base.
#define CSHEADSIZE (offsetof(struct cstr_base, data))

// Requested capacity is clamped to the double of this value.
#define CSHALFSIZE_MAX (((size_t)PTRDIFF_MAX - CSHEADSIZE) >> 1)

// Convenience functions which calculate the address of the beginning of a cstr_base from its data member, and casts it to the required pointer type.
static inline pcsbase get_csbase_addr(cstr cs)
{
  return (pcsbase)(void *)(cs - CSHEADSIZE); // The intermediate cast to void* suppresses the warning that the alignment increases from 1 to sizeof(size_t).
}

static inline pccsbase get_ccsbase_addr(ccstr cs)
{
  return (pccsbase)(const void *)(cs - CSHEADSIZE);
}

// (Re-)allocate the necessary memory, die if the allocation failed.
static pcsbase cs_alloc_or_die(pcsbase ptr, size_t req)
{
  if (ptr != NULL && req == (CSHALFSIZE_MAX << 1) && ptr->capacity == req - 1U)
    CSDIE; // The maximum possible memory has been already allocated. Another call of this function won't return more memory anyway.

  const pcsbase ret = (pcsbase)realloc(ptr, CSHEADSIZE + req);
  if (!ret)
    CSDIE; // Not so nice. But not getting memory anymore means that your OS is close to collapse. Better immediately give it a chance to recover.

  ret->capacity = req - 1U;
  return ret;
}

cstr cs_init(const char *str)
{
  return cs_init_n(str, strlen(str));
}

cstr cs_init_format(const char *format, ...)
{
  va_list args;
  va_start(args, format);
  int len = vsnprintf(NULL, 0U, format, args);
  va_end(args);
  if (len < 0)
    return NULL;

  const pcsbase ptr = cs_alloc_or_die(NULL, len > 127 ? ((size_t)len + 1U) << 1 : (size_t)256U);
  va_start(args, format);
  len = vsnprintf(ptr->data, ptr->capacity + 1U, format, args);
  va_end(args);
  if (len < 0)
  {
    free(ptr);
    return NULL;
  }

  ptr->size = (size_t)len;
  return ptr->data;
}

cstr cs_init_n(const char *seq, size_t cnt)
{
  const size_t inclnull = cstr_add_or_die(CSHALFSIZE_MAX << 1, cnt, 1U),
               req = cnt > 127U ? (inclnull > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : inclnull) << 1 : (size_t)256U;
  const pcsbase ptr = cs_alloc_or_die(NULL, req);
  ptr->size = cnt;
  memcpy(ptr->data + 0, seq, cnt); // +0 to decay the array to a pointer; analysis tools may consider the possibility to violate the declared array boundary (which is intentionally the case)
  *(ptr->data + cnt) = '\0'; // no subscript operator here for the same reason as above; those workarounds are used in other functions, too
  return ptr->data;
}

cstr cs_init_rdentirefile(FILE *stream)
{
  size_t len = 0U;
  if (file_size(stream, &len))
  {
    rewind(stream);
    const size_t inclnull = cstr_add_or_die(CSHALFSIZE_MAX << 1, len, 1U);
    const pcsbase ptr = cs_alloc_or_die(NULL, (len < 256U) ? (size_t)256U : inclnull);
    len = fread(ptr->data + 0, 1U, len, stream);
    *(ptr->data + len) = '\0';
    ptr->size = len;
    return ptr->data;
  }

  return NULL;
}

cstr cs_init_rdfile(FILE *stream, size_t cnt)
{
  // Die if the user tries to read more than the maximum length of a cstr.
  size_t len = 0U, req = (cnt == SIZE_MAX || cnt < 256U) ? (size_t)256U : cstr_add_or_die(CSHALFSIZE_MAX << 1, cnt, 1U);
  pcsbase ptr = cs_alloc_or_die(NULL, req);
  char *cursor = ptr->data;
  for (int read = EOF; len < cnt && (read = fgetc(stream)) != EOF;)
  {
    *cursor++ = (char)read;
    if (++len == req)
    {
      ptr = cs_alloc_or_die(ptr, (req = (req > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : req) << 1));
      cursor = ptr->data + len;
    }
  }

  *cursor = '\0';
  ptr->size = len;
  return ptr->data;
}

cstr cs_init_rdline(FILE *stream)
{
  size_t len = 0U, req = 256U;
  pcsbase ptr = cs_alloc_or_die(NULL, req);
  char *cursor = ptr->data;
  for (int read = EOF; (read = fgetc(stream)) != '\n' && read != EOF;)
  {
    *cursor++ = (char)read;
    if (++len == req)
    {
      ptr = cs_alloc_or_die(ptr, (req = (req > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : req) << 1));
      cursor = ptr->data + len;
    }
  }

  *cursor = '\0';
  ptr->size = len;
  return ptr->data;
}

cstr cs_init_zero(size_t size)
{
  const size_t inclnull = cstr_add_or_die(CSHALFSIZE_MAX << 1, size, 1U),
               req = size > 127U ? (inclnull > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : inclnull) << 1 : (size_t)256U;
  const pcsbase ptr = cs_alloc_or_die(NULL, req);
  ptr->size = 0U;
  *ptr->data = '\0';
  return ptr->data;
}

void cs_free(ccstr cs)
{
  // Work around the ancient quirk in the C lib where the parameter of free() is `void*` rather then `const void*`.
  free(discard_const(get_ccsbase_addr(cs)));
}

size_t cs_max_size(void)
{
  return (CSHALFSIZE_MAX << 1) - 1U;
}

cstr cs_append(cstr *pcs, const char *str)
{
  return cs_append_n(pcs, str, strlen(str));
}

cstr cs_append_n(cstr *pcs, const char *seq, size_t cnt)
{
  if (!cnt)
    return *pcs;

  pcsbase ptr = get_csbase_addr(*pcs);
  const size_t req = cstr_add_or_die(CSHALFSIZE_MAX << 1, ptr->size, cstr_add_or_die(CSHALFSIZE_MAX << 1, cnt, 1U));
  if (req >= ptr->capacity)
    ptr = cs_alloc_or_die(ptr, (req > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : req) << 1);

  memcpy(ptr->data + ptr->size, seq, cnt);
  ptr->size += cnt;
  *(ptr->data + ptr->size) = '\0';
  *pcs = ptr->data;
  return ptr->data;
}

cstr cs_assign(cstr *pcs, const char *str)
{
  return cs_assign_n(pcs, str, strlen(str));
}

cstr cs_assign_n(cstr *pcs, const char *seq, size_t cnt)
{
  pcsbase ptr = get_csbase_addr(*pcs);
  if (cnt > ptr->capacity)
  {
    const size_t inclnull = cstr_add_or_die(CSHALFSIZE_MAX << 1, cnt, 1U);
    ptr = cs_alloc_or_die(ptr, (inclnull > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : inclnull) << 1);
  }

  memcpy(ptr->data + 0, seq, cnt);
  *(ptr->data + cnt) = '\0';
  ptr->size = cnt;
  *pcs = ptr->data;
  return ptr->data;
}

char cs_at(ccstr cs, size_t pos)
{
  const pccsbase ptr = get_ccsbase_addr(cs);
  if (pos >= ptr->size)
    return '\0';

  return cs[pos];
}

size_t cs_capacity(ccstr cs)
{
  const pccsbase ptr = get_ccsbase_addr(cs);
  return ptr->capacity;
}

cstr cs_clear(cstr cs)
{
  const pcsbase cur = get_csbase_addr(cs);
  cur->size = 0U;
  *cs = '\0';
  return cs;
}

bool cs_empty(ccstr cs)
{
  const pccsbase ptr = get_ccsbase_addr(cs);
  return !ptr->size;
}

cstr cs_erase(cstr cs, size_t pos, size_t length)
{
  const pcsbase ptr = get_csbase_addr(cs);
  if (pos >= ptr->size)
    return cs;

  const size_t diff = ptr->size - pos,
               len = (diff < length ? diff : length);

  char *const cursor = cs + pos;
  memmove(cursor, cursor + len, diff - len + 1U);
  ptr->size -= len;
  return cs;
}

cstr cs_fix(cstr *pcs, size_t length, const char *pad, unsigned mode)
{
  pcsbase ptr = get_csbase_addr(*pcs);
  const long long pad_len = pad ? (long long)strlen(pad) : 0LL,
                  diff = (long long)((unsigned long long)length - ptr->size);

  if (!diff || (diff > 0LL && !pad_len))
    return *pcs;

  const long long head_len = (mode & 1 && mode & 2) ? diff >> 1 : ((mode & 1) ? diff : 0LL);

  if (diff < 0LL)
  {
    memmove(*pcs, *pcs - head_len, length);
    *(*pcs + length) = '\0';
    ptr->size = length;
    return *pcs;
  }

  if (length >= ptr->capacity)
  {
    const size_t inclnull = cstr_add_or_die(CSHALFSIZE_MAX << 1, length, 1U);
    ptr = cs_alloc_or_die(ptr, (inclnull > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : inclnull) << 1);
  }

  char *cs_it = ptr->data;
  const char *pad_it = pad,
             *const pad_end = pad + pad_len;

  if (mode & 1)
  {
    memmove(ptr->data + head_len, ptr->data + 0, ptr->size);
    for (char *head_end = cs_it + head_len; cs_it < head_end; ++cs_it)
    {
      *cs_it = *pad_it;
      if (++pad_it == pad_end)
        pad_it = pad;
    }
  }

  cs_it += ptr->size;
  if (mode & 2)
  {
    pad_it = pad;
    for (char *tail_end = cs_it + diff - head_len; cs_it < tail_end; ++cs_it)
    {
      *cs_it = *pad_it;
      if (++pad_it == pad_end)
        pad_it = pad;
    }
  }

  *cs_it = '\0';
  ptr->size = length;
  return ptr->data;
}

cstr cs_insert(cstr *pcs, size_t pos, const char *str)
{
  return cs_insert_n(pcs, pos, str, strlen(str));
}

cstr cs_insert_n(cstr *pcs, size_t pos, const char *seq, size_t cnt)
{
  pcsbase ptr = get_csbase_addr(*pcs);
  if (pos >= ptr->size)
    return cs_append_n(pcs, seq, cnt);

  const size_t req = cstr_add_or_die(CSHALFSIZE_MAX << 1, ptr->size, cstr_add_or_die(CSHALFSIZE_MAX << 1, cnt, 1U));

  if (req >= ptr->capacity)
    ptr = cs_alloc_or_die(ptr, (req > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : req) << 1);

  char *const cursor = ptr->data + pos;
  memmove(cursor + cnt, cursor, ptr->size - pos + 1U);
  memcpy(cursor, seq, cnt);
  ptr->size += cnt;
  *pcs = ptr->data;
  return ptr->data;
}

size_t cs_length(ccstr cs)
{
  return cs_size(cs);
}

char cs_pop_back(cstr cs)
{
  const pcsbase ptr = get_csbase_addr(cs);
  if (!ptr->size)
    return '\0';

  const char ch = cs[--ptr->size];
  cs[ptr->size] = '\0';
  return ch;
}

cstr cs_push_back(cstr *pcs, const char ch)
{
  pcsbase ptr = get_csbase_addr(*pcs);
  const size_t req = cstr_add_or_die(CSHALFSIZE_MAX << 1, ptr->size, 2U);
  if (req >= ptr->capacity)
    ptr = cs_alloc_or_die(ptr, (req > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : req) << 1);

  *(ptr->data + ptr->size++) = ch;
  *(ptr->data + ptr->size) = '\0';
  *pcs = ptr->data;
  return ptr->data;
}

cstr cs_replace(cstr *pcs, size_t pos, size_t length, const char *str)
{
  return cs_replace_n(pcs, pos, length, str, strlen(str));
}

cstr cs_replace_n(cstr *pcs, size_t pos, size_t length, const char *seq, size_t cnt)
{
  pcsbase ptr = get_csbase_addr(*pcs);
  if (pos >= ptr->size)
    return cs_append_n(pcs, seq, cnt);

  const size_t req = cstr_add_or_die(CSHALFSIZE_MAX << 1, ptr->size - (length > ptr->size ? ptr->size : length), cstr_add_or_die(CSHALFSIZE_MAX << 1, cnt, 1U));
  if (req >= ptr->capacity)
    ptr = cs_alloc_or_die(ptr, (req > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : req) << 1);

  *pcs = ptr->data;
  const size_t rem = ptr->size - pos;
  char *const cursor = ptr->data + pos;
  if (rem <= length)
  {
    memcpy(cursor, seq, cnt);
    ptr->size = pos + cnt;
    *(ptr->data + ptr->size) = '\0';
    return ptr->data;
  }

  if (cnt == length)
  {
    memcpy(cursor, seq, cnt);
    return ptr->data;
  }

  memmove(cursor + cnt, cursor + length, rem + 1U);
  memcpy(cursor, seq, cnt);
  ptr->size += cnt - length;
  return ptr->data;
}

cstr cs_reserve(cstr *pcs, size_t capa)
{
  pcsbase ptr = get_csbase_addr(*pcs);
  if (ptr->capacity >= capa)
    return *pcs;

  const size_t inclnull = cstr_add_or_die(CSHALFSIZE_MAX << 1, capa, 1U);
  ptr = cs_alloc_or_die(ptr, (inclnull > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : inclnull) << 1);
  *pcs = ptr->data;
  return ptr->data;
}

cstr cs_resize(cstr *pcs, size_t size, char fill)
{
  pcsbase ptr = get_csbase_addr(*pcs);
  if (ptr->size == size)
    return *pcs;

  if (ptr->size > size)
  {
    (*pcs)[size] = '\0';
    ptr->size = size;
    return *pcs;
  }

  if (size > ptr->capacity)
  {
    const size_t inclnull = cstr_add_or_die(CSHALFSIZE_MAX << 1, size, 1U);
    ptr = cs_alloc_or_die(ptr, (inclnull > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : inclnull) << 1);
  }

  char *cursor = ptr->data + ptr->size;
  memset(cursor, fill, (size_t)(ptr->data + size - cursor));
  cursor[size] = '\0';
  ptr->size = size;
  *pcs = ptr->data;
  return ptr->data;
}

cstr cs_reverse(cstr cs)
{
  pcsbase ptr = get_csbase_addr(cs);
  if (ptr->size > 1U)
    for (char *end = cs + ptr->size - 1U, c = '\0'; cs < end; c = *cs, *cs++ = *end, *end-- = c)
      ;

  return ptr->data;
}

cstr cs_shrink_to_fit(cstr *pcs)
{
  pcsbase ptr = get_csbase_addr(*pcs);
  if (ptr->capacity > ptr->size)
    ptr = cs_alloc_or_die(ptr, ptr->size + 1U);

  *pcs = ptr->data;
  return ptr->data;
}

size_t cs_size(ccstr cs)
{
  const pccsbase ptr = get_ccsbase_addr(cs);
  return ptr->size;
}

cstr cs_substr(ccstr cs, size_t pos, size_t length)
{
  const pccsbase cur = get_ccsbase_addr(cs);
  if (pos > cur->size)
    return cs_init_n("", 0U);

  const size_t diff = cur->size - pos,
               len = diff < length ? diff : length,
               req = len > 127U ? (len + 1U > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : len + 1U) << 1 : (size_t)256U;
  const pcsbase ret = cs_alloc_or_die(NULL, req);
  ret->size = len;
  memcpy(ret->data + 0, cur->data + pos, ret->size);
  *(ret->data + ret->size) = '\0';
  return ret->data;
}

cstr cs_trim(cstr cs, const char *char_set, unsigned mode)
{
  const pcsbase ptr = get_csbase_addr(cs);

  char *it = NULL;
  if (mode & 2)
  {
    for (it = ptr->data + ptr->size; it > ptr->data && strpbrk(it - 1, char_set) == it - 1; --it)
      ;

    *it = '\0';
    ptr->size = (size_t)(it - ptr->data);
  }

  if (!ptr->size)
    return cs;

  if (mode & 1)
  {
#if defined(_MSC_VER)
#  pragma warning(suppress : 6296) // MS wrongly reports that the loop body is only executed once
#endif
    for (it = ptr->data; strpbrk(it, char_set) == it; ++it)
      ;

    if (ptr->data != it)
    {
      ptr->size -= (size_t)(it - ptr->data);
      memmove(ptr->data + 0, it, ptr->size + 1U);
    }
  }

  return cs;
}

cstr cs_update_length(cstr cs, size_t length)
{
  const pcsbase ptr = get_csbase_addr(cs);
  ptr->size = length > ptr->capacity ? ptr->capacity : length;
  cs[ptr->size] = '\0';
  return cs;
}

size_t cs_utf8_count(ccstr cs)
{
  const unsigned char *it = (const unsigned char *)cs;
  size_t cnt = 0U;

  for (; *it != 0U; ++cnt)
    if (u8_next_codepoint(&it) == false)
      return (size_t)0U;

  if (it == (const unsigned char *)(cs + get_ccsbase_addr(cs)->size))
    return cnt;

  return (size_t)0U;
}

cstr cs_utf8_handle_partials(cstr *pcs, u8state_t *pstate)
{
  const unsigned char masks[] = { (unsigned char)0x00, (unsigned char)0xC0, (unsigned char)0xE0, (unsigned char)0xF0 };

  pcsbase ptr = get_csbase_addr(*pcs);
  if (ptr->size == 0U) // escape early because cached partials would be unnecessarily copied into the string and then cached again
    return *pcs;

  if (pstate->length != 0U)
  {
    const size_t ttl = cstr_add_or_die(CSHALFSIZE_MAX << 1, ptr->size, (size_t)pstate->length);
    if (ptr->capacity < ttl)
    {
      const size_t req = cstr_add_or_die(CSHALFSIZE_MAX << 1, ttl, 1U);
      ptr = cs_alloc_or_die(ptr, (req > CSHALFSIZE_MAX ? CSHALFSIZE_MAX : req) << 1);
    }

    memmove(ptr->data + pstate->length, ptr->data + 0, ptr->size + 1U);
    memcpy(ptr->data + 0, pstate->cache + 0, (size_t)pstate->length);
    ptr->size += pstate->length;
    pstate->length = 0U;
  }

  unsigned char *back_it = (unsigned char *)(ptr->data + ptr->size);
  if ((*(back_it - 1) & (unsigned char)0x80) == (unsigned char)0x80)
  {
    const size_t stop_len = ptr->size < 3U ? ptr->size : (size_t)3U;
    for (unsigned sequence_len = 1U; sequence_len <= stop_len; ++sequence_len)
    {
      --back_it;
      if ((*back_it & (unsigned char)0xC0) == (unsigned char)0xC0)
      {
        if ((*back_it & masks[sequence_len]) != masks[sequence_len - 1U])
        {
          pstate->length = sequence_len;
          memcpy(pstate->cache + 0, back_it, sequence_len);
          *back_it = (unsigned char)0x00;
          ptr->size -= sequence_len;
        }

        break;
      }
    }
  }

  *pcs = ptr->data;
  return ptr->data;
}

bool cs_utf8_hasbom(ccstr cs)
{
  if (*cs == '\xEF' && cs[1] == '\xBF' && cs[2] == '\xBD')
    return true;

  return false;
}

size_t cs_utf8_idx_of_nth_codepoint(ccstr cs, size_t pos)
{
  const unsigned char *it = (const unsigned char *)cs;
  size_t idx = 0U;

  for (; *it != 0U && pos > idx; ++idx)
    if (u8_next_codepoint(&it) == false)
      return (size_t)0U;

  if (pos == idx)
    return (size_t)(it - (const unsigned char *)cs);

  return (size_t)0U;
}

cstr cs_utf8_replace_invalids(cstr *pcs, const char *rep)
{
  const char repl_char[] = { '\xEF', '\xBF', '\xBD' };
  pcsbase ptr = get_csbase_addr(*pcs);
  if (ptr->size == 0U)
    return *pcs;

  const char *repl = rep == NULL ? repl_char : rep;
  const size_t repl_len = rep == NULL ? sizeof(repl_char) : strlen(rep),
               factor = (repl_len == 0U ? (size_t)1U : repl_len);
  if (ptr->size > (CSHALFSIZE_MAX << 1) / factor) // in the worst case every code unit will be replaced
    CSDIE;

  const size_t req = cstr_add_or_die((CSHALFSIZE_MAX << 1), ptr->size * factor, 1U);
  size_t delta = 0U;
  pcsbase ret = cs_alloc_or_die(NULL, req > (CSHALFSIZE_MAX << 1) ? (CSHALFSIZE_MAX << 1) : req);
  const unsigned char *it = (unsigned char *)(*pcs),
                      *last = it;
  char *it_ret = ret->data;
  for (const unsigned char *tmp = it; *it != 0U; it = tmp)
  {
    if (u8_next_codepoint(&tmp))
      delta += (size_t)(tmp - it);
    else
    {
      if (delta != 0U)
      {
        memcpy(it_ret, last, delta);
        it_ret += delta;
        delta = 0U;
      }

      last = tmp;
      if (repl_len != 0U)
      {
        memcpy(it_ret, repl, repl_len);
        it_ret += repl_len;
      }
    }
  }

  if (delta != 0U)
  {
    memcpy(it_ret, last, delta);
    it_ret += delta;
  }

  free(ptr);
  *it_ret = '\0';
  ret->size = (size_t)(it_ret - ret->data);
  *pcs = ret->data;
  return ret->data;
}

#undef CSHEADSIZE
#undef CSHALFSIZE_MAX

//////////////////////////////////////////////////////////
// vvv  based on wchar_t  vvv

struct cstrw_base
{
  size_t capacity;
  size_t size;
  wchar_t data[sizeof(size_t) / sizeof(wchar_t)];
};

typedef struct cstrw_base *pcswbase;
typedef const struct cstrw_base *pccswbase;

// Distance in bytes of the data member from the beginning of cstr_base.
#define CSWHEADSIZE (offsetof(struct cstrw_base, data))

// Requested capacity is clamped to the double of this value.
#define CSWHALFSIZE_MAX (((size_t)PTRDIFF_MAX / sizeof(wchar_t) - CSWHEADSIZE) >> 1)

static inline pcswbase get_cswbase_addr(cstrw csw)
{
  return (pcswbase)(void *)((unsigned char *)csw - CSWHEADSIZE);
}

static inline pccswbase get_ccswbase_addr(ccstrw csw)
{
  return (pccswbase)(const void *)((const unsigned char *)csw - CSWHEADSIZE);
}

static pcswbase csw_alloc_or_die(pcswbase ptr, size_t req)
{
  if (ptr != NULL && req == (CSWHALFSIZE_MAX << 1) && ptr->capacity == req - 1U)
    CSDIE; // The maximum possible memory has been already allocated. Another call of this function won't return more memory anyway.

  const pcswbase ret = (pcswbase)realloc(ptr, CSWHEADSIZE + req * sizeof(wchar_t));
  if (!ret)
    CSDIE; // Not so nice. But not getting memory anymore means that your OS is close to collapse. Better immediately give it a chance to recover.

  ret->capacity = req - 1U;
  return ret;
}

cstrw csw_init(const wchar_t *strw)
{
  return csw_init_n(strw, wcslen(strw));
}

cstrw csw_init_format(const wchar_t *formatw, ...)
{
  va_list args;
#if defined(_WIN32) || defined(__CYGWIN__) // Windows default, including MinGW and Cygwin
  va_start(args, formatw);

#  if defined(_MSC_VER)
#    pragma warning(push)
#    pragma warning(disable : 4996) // MS reports _vsnwprintf to be deprecated, but that's the function which returns the necessary length
#  endif
  int len = _vsnwprintf(NULL, 0U, formatw, args);
#  if defined(_MSC_VER)
#    pragma warning(pop) // C4996
#  endif

  va_end(args);
  if (len < 0)
    return NULL;

  const pcswbase ptr = csw_alloc_or_die(NULL, len > 127 ? ((size_t)len + 1U) << 1 : (size_t)256U);
  va_start(args, formatw);
  len = _vsnwprintf_s(ptr->data, ptr->capacity + 1U, _TRUNCATE, formatw, args);
  va_end(args);
  if (len < 0)
  {
    free(ptr);
    return NULL;
  }
#else // work around the non-standard _vsnwprintf using vswprintf
  const size_t size_max = (size_t)(INT_MAX) + 1U;
  pcswbase ptr = NULL;
  int len = 0;
  for (size_t req = (size_t)127U;;)
  {
    if (req >= size_max)
      return NULL;

    req = (req + 1U) << 1;
    if (req > size_max)
      req = size_max;

    ptr = csw_alloc_or_die(NULL, req);
    va_start(args, formatw);
    len = vswprintf(ptr->data, ptr->capacity + 1U, formatw, args);
    va_end(args);
    if (len >= 0)
      break;

    free(ptr);
  }
#endif

  ptr->size = (size_t)len;
  return ptr->data;
}

cstrw csw_init_n(const wchar_t *seqw, size_t cnt)
{
  const size_t inclnull = cstr_add_or_die(CSWHALFSIZE_MAX << 1, cnt, 1U),
               req = cnt > 127U ? (inclnull > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : inclnull) << 1 : (size_t)256U;
  const pcswbase ptr = csw_alloc_or_die(NULL, req);
  ptr->size = cnt;
  memcpy(ptr->data + 0, seqw, cnt * sizeof(wchar_t));
  *(ptr->data + cnt) = L'\0';
  return ptr->data;
}

cstrw csw_init_rdentirefile(FILE *stream)
{
  size_t len = 0U;
  if (file_size(stream, &len))
  {
    rewind(stream);
    len /= sizeof(wchar_t);
    const size_t inclnull = cstr_add_or_die(CSWHALFSIZE_MAX << 1, len, 1U);
    const pcswbase ptr = csw_alloc_or_die(NULL, (len < 256U) ? (size_t)256U : inclnull);
    len = fread(ptr->data + 0, sizeof(wchar_t), len, stream);
    *(ptr->data + len) = L'\0';
    ptr->size = len;
    return ptr->data;
  }

  return NULL;
}

cstrw csw_init_rdfile(FILE *stream, size_t cnt)
{
  // Die if the user tries to read more than the maximum length of a cstrw.
  size_t len = 0U, req = (cnt == SIZE_MAX || cnt < 256U) ? (size_t)256U : cstr_add_or_die(CSWHALFSIZE_MAX << 1, cnt, 1U);
  pcswbase ptr = csw_alloc_or_die(NULL, req);
  wchar_t *cursor = ptr->data;
  for (wint_t read = WEOF; len < cnt && (read = fgetwc(stream)) != WEOF;)
  {
    *cursor++ = (wchar_t)read;
    if (++len == req)
    {
      ptr = csw_alloc_or_die(ptr, (req = (req > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : req) << 1));
      cursor = ptr->data + len;
    }
  }

  *cursor = L'\0';
  ptr->size = len;
  return ptr->data;
}

cstrw csw_init_rdline(FILE *stream)
{
  size_t len = 0U, req = 256U;
  pcswbase ptr = csw_alloc_or_die(NULL, req);
  wchar_t *cursor = ptr->data;
  for (wint_t read = WEOF; (read = fgetwc(stream)) != L'\n' && read != WEOF;)
  {
    *cursor++ = (wchar_t)read;
    if (++len == req)
    {
      ptr = csw_alloc_or_die(ptr, (req = (req > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : req) << 1));
      cursor = ptr->data + len;
    }
  }

  *cursor = L'\0';
  ptr->size = len;
  return ptr->data;
}

cstrw csw_init_zero(size_t size)
{
  const size_t inclnull = cstr_add_or_die(CSWHALFSIZE_MAX << 1, size, 1U),
               req = size > 127U ? (inclnull > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : inclnull) << 1 : (size_t)256U;
  const pcswbase ptr = csw_alloc_or_die(NULL, req);
  ptr->size = 0U;
  *ptr->data = L'\0';
  return ptr->data;
}

void csw_free(ccstrw csw)
{
  free(discard_const(get_ccswbase_addr(csw)));
}

size_t csw_max_size(void)
{
  return (CSWHALFSIZE_MAX << 1) - 1U;
}

cstrw csw_append(cstrw *pcsw, const wchar_t *strw)
{
  return csw_append_n(pcsw, strw, wcslen(strw));
}

cstrw csw_append_n(cstrw *pcsw, const wchar_t *seqw, size_t cnt)
{
  if (!cnt)
    return *pcsw;

  pcswbase ptr = get_cswbase_addr(*pcsw);
  const size_t req = cstr_add_or_die(CSWHALFSIZE_MAX << 1, ptr->size, cstr_add_or_die(CSWHALFSIZE_MAX << 1, cnt, 1U));
  if (req >= ptr->capacity)
    ptr = csw_alloc_or_die(ptr, (req > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : req) << 1);

  memcpy(ptr->data + ptr->size, seqw, cnt * sizeof(wchar_t));
  ptr->size += cnt;
  *(ptr->data + ptr->size) = L'\0';
  *pcsw = ptr->data;
  return ptr->data;
}

cstrw csw_assign(cstrw *pcsw, const wchar_t *strw)
{
  return csw_assign_n(pcsw, strw, wcslen(strw));
}

cstrw csw_assign_n(cstrw *pcsw, const wchar_t *seqw, size_t cnt)
{
  pcswbase ptr = get_cswbase_addr(*pcsw);
  if (cnt > ptr->capacity)
  {
    const size_t inclnull = cstr_add_or_die(CSWHALFSIZE_MAX << 1, cnt, 1U);
    ptr = csw_alloc_or_die(ptr, (inclnull > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : inclnull) << 1);
  }

  memcpy(ptr->data + 0, seqw, cnt * sizeof(wchar_t));
  *(ptr->data + cnt) = L'\0';
  ptr->size = cnt;
  *pcsw = ptr->data;
  return ptr->data;
}

wchar_t csw_at(ccstrw csw, size_t pos)
{
  const pccswbase ptr = get_ccswbase_addr(csw);
  if (pos >= ptr->size)
    return L'\0';

  return csw[pos];
}

size_t csw_capacity(ccstrw csw)
{
  const pccswbase ptr = get_ccswbase_addr(csw);
  return ptr->capacity;
}

cstrw csw_clear(cstrw csw)
{
  const pcswbase cur = get_cswbase_addr(csw);
  cur->size = 0U;
  *csw = L'\0';
  return csw;
}

bool csw_empty(ccstrw csw)
{
  const pccswbase ptr = get_ccswbase_addr(csw);
  return !ptr->size;
}

cstrw csw_erase(cstrw csw, size_t pos, size_t length)
{
  const pcswbase ptr = get_cswbase_addr(csw);
  if (pos >= ptr->size)
    return csw;

  const size_t diff = ptr->size - pos,
               len = (diff < length ? diff : length);

  wchar_t *const cursor = csw + pos;
  memmove(cursor, cursor + len, (diff - len + 1U) * sizeof(wchar_t));
  ptr->size -= len;
  return csw;
}

cstrw csw_fix(cstrw *pcsw, size_t length, const wchar_t *padw, unsigned mode)
{
  pcswbase ptr = get_cswbase_addr(*pcsw);
  const long long pad_len = padw ? (long long)wcslen(padw) : 0LL,
                  diff = (long long)((unsigned long long)length - ptr->size);

  if (!diff || (diff > 0LL && !pad_len))
    return *pcsw;

  const long long head_len = (mode & 1 && mode & 2) ? diff >> 1 : ((mode & 1) ? diff : 0LL);

  if (diff < 0LL)
  {
    memmove(*pcsw, *pcsw - head_len, length * sizeof(wchar_t));
    *(*pcsw + length) = L'\0';
    ptr->size = length;
    return *pcsw;
  }

  if (length >= ptr->capacity)
  {
    const size_t inclnull = cstr_add_or_die(CSWHALFSIZE_MAX << 1, length, 1U);
    ptr = csw_alloc_or_die(ptr, (inclnull > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : inclnull) << 1);
  }

  wchar_t *cs_it = ptr->data;
  const wchar_t *pad_it = padw,
                *const pad_end = padw + pad_len;

  if (mode & 1)
  {
    memmove(ptr->data + head_len, ptr->data + 0, ptr->size * sizeof(wchar_t));
    for (wchar_t *head_end = cs_it + head_len; cs_it < head_end; ++cs_it)
    {
      *cs_it = *pad_it;
      if (++pad_it == pad_end)
        pad_it = padw;
    }
  }

  cs_it += ptr->size;
  if (mode & 2)
  {
    pad_it = padw;
    for (wchar_t *tail_end = cs_it + diff - head_len; cs_it < tail_end; ++cs_it)
    {
      *cs_it = *pad_it;
      if (++pad_it == pad_end)
        pad_it = padw;
    }
  }

  *cs_it = L'\0';
  ptr->size = length;
  return ptr->data;
}

cstrw csw_insert(cstrw *pcsw, size_t pos, const wchar_t *strw)
{
  return csw_insert_n(pcsw, pos, strw, wcslen(strw));
}

cstrw csw_insert_n(cstrw *pcsw, size_t pos, const wchar_t *seqw, size_t cnt)
{
  pcswbase ptr = get_cswbase_addr(*pcsw);
  if (pos >= ptr->size)
    return csw_append_n(pcsw, seqw, cnt);

  const size_t req = cstr_add_or_die(CSWHALFSIZE_MAX << 1, ptr->size, cstr_add_or_die(CSWHALFSIZE_MAX << 1, cnt, 1U));

  if (req >= ptr->capacity)
    ptr = csw_alloc_or_die(ptr, (req > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : req) << 1);

  wchar_t *const cursor = ptr->data + pos;
  memmove(cursor + cnt, cursor, (ptr->size - pos + 1U) * sizeof(wchar_t));
  memcpy(cursor, seqw, cnt * sizeof(wchar_t));
  ptr->size += cnt;
  *pcsw = ptr->data;
  return ptr->data;
}

size_t csw_length(ccstrw csw)
{
  return csw_size(csw);
}

wchar_t csw_pop_back(cstrw csw)
{
  const pcswbase ptr = get_cswbase_addr(csw);
  if (!ptr->size)
    return L'\0';

  const wchar_t ch = csw[--ptr->size];
  csw[ptr->size] = L'\0';
  return ch;
}

cstrw csw_push_back(cstrw *pcsw, const wchar_t chw)
{
  pcswbase ptr = get_cswbase_addr(*pcsw);
  const size_t req = cstr_add_or_die(CSWHALFSIZE_MAX << 1, ptr->size, 2U);
  if (req >= ptr->capacity)
    ptr = csw_alloc_or_die(ptr, (req > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : req) << 1);

  *(ptr->data + ptr->size++) = chw;
  *(ptr->data + ptr->size) = L'\0';
  *pcsw = ptr->data;
  return ptr->data;
}

cstrw csw_replace(cstrw *pcsw, size_t pos, size_t length, const wchar_t *strw)
{
  return csw_replace_n(pcsw, pos, length, strw, wcslen(strw));
}

cstrw csw_replace_n(cstrw *pcsw, size_t pos, size_t length, const wchar_t *seqw, size_t cnt)
{
  pcswbase ptr = get_cswbase_addr(*pcsw);
  if (pos >= ptr->size)
    return csw_append_n(pcsw, seqw, cnt);

  const size_t req = cstr_add_or_die(CSWHALFSIZE_MAX << 1, ptr->size - (length > ptr->size ? ptr->size : length), cstr_add_or_die(CSWHALFSIZE_MAX << 1, cnt, 1U));
  if (req >= ptr->capacity)
    ptr = csw_alloc_or_die(ptr, (req > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : req) << 1);

  *pcsw = ptr->data;
  const size_t rem = ptr->size - pos;
  wchar_t *const cursor = ptr->data + pos;
  if (rem <= length)
  {
    memcpy(cursor, seqw, cnt * sizeof(wchar_t));
    ptr->size = pos + cnt;
    *(ptr->data + ptr->size) = L'\0';
    return ptr->data;
  }

  if (cnt == length)
  {
    memcpy(cursor, seqw, cnt * sizeof(wchar_t));
    return ptr->data;
  }

  memmove(cursor + cnt, cursor + length, (rem + 1U) * sizeof(wchar_t));
  memcpy(cursor, seqw, cnt * sizeof(wchar_t));
  ptr->size += cnt - length;
  return ptr->data;
}

cstrw csw_reserve(cstrw *pcsw, size_t capa)
{
  pcswbase ptr = get_cswbase_addr(*pcsw);
  if (ptr->capacity >= capa)
    return *pcsw;

  const size_t inclnull = cstr_add_or_die(CSWHALFSIZE_MAX << 1, capa, 1U);
  ptr = csw_alloc_or_die(ptr, (inclnull > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : inclnull) << 1);
  *pcsw = ptr->data;
  return ptr->data;
}

cstrw csw_resize(cstrw *pcsw, size_t size, wchar_t fillw)
{
  pcswbase ptr = get_cswbase_addr(*pcsw);
  if (ptr->size == size)
    return *pcsw;

  if (ptr->size > size)
  {
    (*pcsw)[size] = L'\0';
    ptr->size = size;
    return *pcsw;
  }

  if (size > ptr->capacity)
  {
    const size_t inclnull = cstr_add_or_die(CSWHALFSIZE_MAX << 1, size, 1U);
    ptr = csw_alloc_or_die(ptr, (inclnull > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : inclnull) << 1);
  }

  wchar_t *cursor = ptr->data + ptr->size;
  for (wchar_t *end = ptr->data + size; cursor < end; *cursor++ = fillw)
    ;

  *cursor = L'\0';
  ptr->size = size;
  *pcsw = ptr->data;
  return ptr->data;
}

cstrw csw_reverse(cstrw csw)
{
  pcswbase ptr = get_cswbase_addr(csw);
  if (ptr->size > 1U)
    for (wchar_t *end = csw + ptr->size - 1U, c = L'\0'; csw < end; c = *csw, *csw++ = *end, *end-- = c)
      ;

  return ptr->data;
}

cstrw csw_shrink_to_fit(cstrw *pcsw)
{
  pcswbase ptr = get_cswbase_addr(*pcsw);
  if (ptr->capacity > ptr->size)
    ptr = csw_alloc_or_die(ptr, ptr->size + 1U);

  *pcsw = ptr->data;
  return ptr->data;
}

size_t csw_size(ccstrw csw)
{
  const pccswbase ptr = get_ccswbase_addr(csw);
  return ptr->size;
}

cstrw csw_substr(ccstrw csw, size_t pos, size_t length)
{
  const pccswbase cur = get_ccswbase_addr(csw);
  if (pos > cur->size)
    return csw_init_n(L"", 0U);

  const size_t diff = cur->size - pos,
               len = diff < length ? diff : length,
               req = len > 127U ? (len + 1U > CSWHALFSIZE_MAX ? CSWHALFSIZE_MAX : len + 1U) << 1 : (size_t)256U;
  const pcswbase ret = csw_alloc_or_die(NULL, req);
  ret->size = len;
  memcpy(ret->data + 0, cur->data + pos, ret->size * sizeof(wchar_t));
  *(ret->data + ret->size) = L'\0';
  return ret->data;
}

cstrw csw_trim(cstrw csw, const wchar_t *char_setw, unsigned mode)
{
  const pcswbase ptr = get_cswbase_addr(csw);

  wchar_t *it = NULL;
  if (mode & 2)
  {
    for (it = ptr->data + ptr->size; it > ptr->data && wcspbrk(it - 1, char_setw) == it - 1; --it)
      ;

    *it = L'\0';
    ptr->size = (size_t)(it - ptr->data);
  }

  if (!ptr->size)
    return csw;

  if (mode & 1)
  {
#if defined(_MSC_VER)
#  pragma warning(suppress : 6296) // MS wrongly reports that the loop body is only executed once
#endif
    for (it = ptr->data; wcspbrk(it, char_setw) == it; ++it)
      ;

    if (ptr->data != it)
    {
      ptr->size -= (size_t)(it - ptr->data);
      memmove(ptr->data + 0, it, (ptr->size + 1U) * sizeof(wchar_t));
    }
  }

  return csw;
}

cstrw csw_update_length(cstrw csw, size_t length)
{
  const pcswbase ptr = get_cswbase_addr(csw);
  ptr->size = length > ptr->capacity ? ptr->capacity : length;
  csw[ptr->size] = L'\0';
  return csw;
}

#if defined(_MSC_VER) && _MSC_VER >= 1914
#  pragma warning(pop) // C5045 Spectre mitigation
#endif
