/*
Copyright (c) 2020 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef CSTR_H_INCLUDED__
#define CSTR_H_INCLUDED__

#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <wchar.h>

//////////////////////////////////////////////////////////
// vvv  based on char  vvv

/** \brief  Type to indicate that the pointer is a cstr.
 *  Besides of that, it is compatible with a pointer to char,
 *  and it behaves like a nullterminated C-string. */
typedef char *cstr;

/** \brief  Type to indicate that the pointer is a cstr
 *  with immutable content. */
typedef const char *ccstr;

/** \brief  Structure type to capture incomplete UTF-8
 *  codepoints. */
typedef struct tag_u8state
{
  unsigned length;
  char cache[4];
} u8state_t;

//////////////////////////////////////////////////////////

/** \brief  Functions to be used for both C-strings and
 *  binary content are marked with the BIN macro.
 *  Even if the library is primarily made for strings
 *  which are compatible with nullterminated C-strings, the
 *  cstr type is capable to store binary data as well. */
#define BIN

//////////////////////////////////////////////////////////

/** \brief  Create and initialize a cstr from a nullterminated string.
 * \param  str  Nullterminated string.
 * \return  Newly allocated cstr containing the copied str. */
cstr cs_init(const char *str);

/** \brief  Create and initialize a cstr from formatted data.
 * \param  format  Format specification. (supported specifiers like for printf)
 * \param  ...     Arguments.
 * \return  Newly allocated cstr containing the formatted data. NULL if the function fails. */
cstr cs_init_format(const char *format, ...)
#if defined(__GNUC__) || defined(__clang__)
  __attribute__((__format__(printf, 1, 2))) __attribute__((__nonnull__(1)))
#endif
  ;

/** \brief  Create and initialize a cstr from a string of characters.
 * \param  seq  String of characters.
 * \param  cnt  First number of characters of seq to be used.
 * \return  Newly allocated cstr containing the copied substring. */
BIN cstr cs_init_n(const char *seq, size_t cnt);

/** \brief  Create and initialize a cstr from the entire file content.
 * \param  stream  File stream to be read. Must not be `stdin`.
 *                 and should have been opened in binary mode.
 * \return  Newly allocated cstr containing the copied characters.
 *          NULL if the function fails to determine the file size. */
BIN cstr cs_init_rdentirefile(FILE *stream);

/** \brief  Create and initialize a cstr from characters read out of a file.
 * \param  stream  File stream to be read. Can be `stdin` if cnt is not -1
 *                 when redirected from keyboard.
 * \param  cnt     First number of characters of stream to be used.
 *                 Pass -1 to read until EOF.
 *                 Note that cnt begins at the current position of the file cursor.
 * \return  Newly allocated cstr containing the copied characters. */
BIN cstr cs_init_rdfile(FILE *stream, size_t cnt);

/** \brief  Create and initialize a cstr from a line of characters read out of a file.
 *          NOTE: This function is able to read binary data but since it stops reading
 *          if line feed was found it may not make much sense in most cases and should
 *          be avoided.
 * \param  stream  File stream to be read until the first occurrence of line feed or EOF.
 *                 Can be `stdin`.
 * \return  Newly allocated cstr containing the copied characters without the line feed
 *          at the end of the line. */
cstr cs_init_rdline(FILE *stream);

/** \brief  Create a zero-length cstr with at least 'size' + 1 bytes storage capacity.
 * \param  size  Expected length of the string to be stored.
 * \return  Newly allocated zero-length cstr. */
BIN cstr cs_init_zero(size_t size);

/** \brief  Delete the cstr and release memory back to the operating system.
 * \param  cs  The cstr to be destroyed. */
BIN void cs_free(ccstr cs);

/** \brief  Get the maximum number of chars that technically can be saved into a cstr.
 * \return  Maximum number of chars, not including the terminating null. */
size_t cs_max_size(void);

//////////////////////////////////////////////////////////

/** \brief  Add data from a nullterminated string at the end of a cstr.
 * \param  pcs  Referenced cstr where new data is appended.
 * \param  str  Nullterminated string.
 * \return  The cstr containing the updated data. */
cstr cs_append(cstr *pcs, const char *str);

/** \brief  Add data from a string of characters at the end of a cstr.
 * \param  pcs  Referenced cstr where new data is appended.
 * \param  seq  String of characters.
 * \param  cnt  First number of characters of seq to be appended.
 * \return  The cstr containing the updated data. */
BIN cstr cs_append_n(cstr *pcs, const char *seq, size_t cnt);

/** \brief  Assign data from a nullterminated string to the cstr, replace its current contents.
 * \param  pcs  Referenced cstr where the data is assigned.
 * \param  str  Nullterminated string.
 * \return  The cstr containing the new data. */
cstr cs_assign(cstr *pcs, const char *str);

/** \brief  Assign data from a string of characters to the cstr, replace its current contents.
 * \param  pcs  Referenced cstr where the data is assigned.
 * \param  seq  String of characters.
 * \param  cnt  First number of characters of seq to be used.
 * \return  The cstr containing the new data. */
BIN cstr cs_assign_n(cstr *pcs, const char *seq, size_t cnt);

/** \brief  Get a character in a cstr from the specified position.
 * \param  cs   The cstr.
 * \param  pos  Zero-based position of a character within the cstr, clamped to its length.
 * \return  The character at the specified position, or '\0' if out of range. */
BIN char cs_at(ccstr cs, size_t pos);

/** \brief  Get the capacity of the allocated character buffer - 1 which is reserved for the terminating null.
 * \param  cs  The cstr.
 * \return  The size of the storage capacity allocated for the cstr. */
BIN size_t cs_capacity(ccstr cs);

/** \brief  Erase the contents of the cstr, which becomes a zero-length string.
 * \param  cs  The cstr.
 * \return  The updated cstr. */
BIN cstr cs_clear(cstr cs);

/** \brief  Test if the cstr has zero length.
 * \param  cs  The cstr.
 * \return  true if the string length is 0, false otherwise. */
BIN bool cs_empty(ccstr cs);

/** \brief  Erase a part of the cstr, reducing its length.
 * \param  cs      The cstr.
 * \param  pos     Zero-based position in the cstr where the erased portion begins.
 *                 If pos is out of range, the cstr is unchanged.
 * \param  length  Length of the portion to be erased.
 *                 If the specified span ends out of range, only the valid portion is erased.
 * \return  The updated cstr. */
BIN cstr cs_erase(cstr cs, size_t pos, size_t length);

/** \brief  Update the cstr to a fixed length by either padding or shortening.
 * \param  pcs     Referenced cstr.
 * \param  length  New length of the cstr.
 *                 If length <= the current length of the cstr, `pad` is ignored and might be NULL.
 * \param  pad     Nullterminated string containing the characters to be padded.
 * \param  mode    Flags which specify where the cstr is to be padded or shortened.
 *                   1 at the begin of the cstr
 *                   2 at the end of the cstr
 *                 Their combination (1|2) leads to a centered alignment.
 * \return  The cstr containing the updated data. */
cstr cs_fix(cstr *pcs, size_t length, const char *pad, unsigned mode);

/** \brief  Insert additional characters into the cstr right before the character indicated by pos.
 * \param  pcs  Referenced cstr where the data is inserted.
 * \param  pos  Zero-based position in the cstr where the characters are inserted.
 *              If pos is out of range, the characters are appended.
 * \param  str  Nullterminated string containing the characters to be inserted.
 * \return  The cstr containing the updated data. */
cstr cs_insert(cstr *pcs, size_t pos, const char *str);

/** \brief  Insert additional characters into the cstr right before the character indicated by pos.
 * \param  pcs  Referenced cstr where the data is inserted.
 * \param  pos  Zero-based position in the cstr where the characters are inserted.
 *              If pos is out of range, the characters are appended.
 * \param  seq  String of characters.
 * \param  cnt  First number of characters of seq to be inserted.
 * \return  The cstr containing the updated data. */
BIN cstr cs_insert_n(cstr *pcs, size_t pos, const char *seq, size_t cnt);

/** \brief  Get the length of the cstr.
 * \param  cs  The cstr.
 * \return  The length of the cstr, without the terminating null-character. */
BIN size_t cs_length(ccstr cs);

/** \brief  Erases the last character of the cstr.
 * \param  cs  The cstr.
 * \return  The character removed from the string end, or '\0' if the cstr has zero length. */
BIN char cs_pop_back(cstr cs);

/** \brief  Append a character to the end of a cstr.
 * \param  pcs  Referenced cstr where the character is appended.
 * \param  ch   Character to be appended.
 * \return  The cstr containing the updated data. */
BIN cstr cs_push_back(cstr *pcs, const char ch);

/** \brief  Replace the portion of the cstr that begins at character 'pos' and spans 'length' characters.
 * \param  pcs     Referenced cstr where data is replaced.
 * \param  pos     Zero-based position in the cstr where the portion begins.
 *                 If pos is out of range, the replacement string appended.
 * \param  length  Length of the portion to be replaced.
 *                 If the portion ends out of range, the cstr will be extended.
 * \param  str     Nullterminated string containing the characters the portion is replaced with.
 * \return  The cstr containing the updated data. */
cstr cs_replace(cstr *pcs, size_t pos, size_t length, const char *str);

/** \brief  Replace the portion of the cstr that begins at character 'pos' and spans 'length' characters.
 * \param  pcs     Referenced cstr where data is replaced.
 * \param  pos     Zero-based position in the cstr where the portion begins.
 *                 If pos is out of range, the replacement string appended.
 * \param  length  Length of the portion to be replaced.
 *                 If the portion ends out of range, the cstr will be extended.
 * \param  seq     String containing the characters the portion is replaced with.
 * \param  cnt     First number of characters of seq to be used.
 * \return  The cstr containing the updated data. */
BIN cstr cs_replace_n(cstr *pcs, size_t pos, size_t length, const char *seq, size_t cnt);

/** \brief  Allocate at least the requested capacity.
 * \param  pcs   Referenced cstr.
 * \param  capa  Minimum capacity of the cstr.
 * \return  The updated cstr. */
BIN cstr cs_reserve(cstr *pcs, size_t capa);

/** \brief  Resize the cstr to a length of 'size' characters.
 * \param  pcs   Referenced cstr to be resized.
 * \param  size  New size of the cstr.
 * \param  fill  Character used to fill the new character space added to the cstr.
 *               If the new size is smaller than the current size, fill is ignored.
 * \return  The updated cstr. */
BIN cstr cs_resize(cstr *pcs, size_t size, char fill);

/** \brief  Reverse the character order in the specified cstr.
 * \param  cs  The cstr.
 * \return  The updated cstr. */
BIN cstr cs_reverse(cstr cs);

/** \brief  Requests the cstr to reduce its capacity to fit its size.
 * \param  pcs  Referenced cstr.
 * \return  The updated cstr. */
BIN cstr cs_shrink_to_fit(cstr *pcs);

/** \brief  Get the length of the cstr.
 * \param  cs  The cstr.
 * \return  The length of the cstr, without the terminating null-character. */
BIN size_t cs_size(ccstr cs);

/** \brief  Newly construct a cstr with its value initialized to a copy of a substring of 'cs'.
 * \param  cs      The cstr.
 * \param  pos     Zero-based position in the cstr where the substring begins.
 *                 If pos is out of range, a zero-length cstr is returned.
 * \param  length  Length of the portion to be copied.
 *                 If the specified span ends out of range, the substring contains only the valid portion.
 * \return  The new cstr containing the copied substring. */
BIN cstr cs_substr(ccstr cs, size_t pos, size_t length);

/** \brief  Remove specified prepended and/or appended characters from a cstr.
 * \param  cs        The cstr.
 * \param  char_set  Nullterminated string of characters which are to be removed.
 * \param  mode      Flags which specify where the characters are to be removed.
 *                     1 to remove prepended characters
 *                     2 to remove appended characters
 * \return  The updated cstr. */
cstr cs_trim(cstr cs, const char *char_set, unsigned mode);

/** \brief  Force the cstr to update its size property to the specified length.
 *          Alteration of a cstr outside of the library is supported as long as this does neither violate
 *          the boundaries of the allocated storage capacity (see cs_capacity), nor its memory address.
 *          In those cases it's necessary to adjust the internal size property of the cstr accordingly.
 * \param  cs      The cstr.
 * \param  length  Actual length of the cstr.
 * \return  The updated cstr with a terminating null set. */
BIN cstr cs_update_length(cstr cs, size_t length);

/** \brief  Count the number of codepoints in a UTF-8-encoded cstr.
 *          This function validates the whole string and returns 0 if invalid characters
 *          will be found.
 * \param  cs  The cstr.
 * \return  The number of codepoints the cstr contains. If the size of the cstr is greater
 *          than 0 and the function returns 0 the cstr contains invalid UTF-8. */
size_t cs_utf8_count(ccstr cs);

/** \brief  Complete partial UTF-8 codepoints at the string boundaries.
 *          If the u8state_t variable contains cached partials from the previous call, the
 *          function will prepend them to the cstr referenced by pcs.
 *          If partials are found at the end of the cstr, they will be removed from the string
 *          and captured in the u8state_t variable referenced by pstate.
 *          This function does not validate the string.
 * \param  pcs     Referenced cstr.
 * \param  pstate  Referenced u8state_t variable.
 *                 Note: The u8state_t variable has to be zero-initialized before it is
 *                 used for the first time.
 * \return  The updated cstr. */
cstr cs_utf8_handle_partials(cstr *pcs, u8state_t *pstate);

/** \brief  Determine whether or not a UTF-8-encoded string has a leading Byte Order Mark.
 *          This function does not validate the string.
 * \param  cs  A cstr or any nullterminated string.
 * \return  true if the first three characters are 0xEF 0xBF 0xBD. false otherwise. */
bool cs_utf8_hasbom(ccstr cs);

/** \brief  Determine the number of characters (code units) between the beginning
 *          of the string and the codepoint specified by pos.
 *          The first part of the string is validated and the function returns 0 if
 *          invalid characters will be found.
 * \param  cs   A cstr or any nullterminated string.
 * \param  pos  Zero-based position of the codepoint in the string.
 * \return  The offset (in chars) from the beginning of the string. If the size of cs
 *          is greater than 0 and the function returns 0 then cs contains either invalid
 *          UTF-8 or pos is out of range. */
size_t cs_utf8_idx_of_nth_codepoint(ccstr cs, size_t pos);

/** \brief  Replace every code unit belonging to an invalid UTF-8 code point to make
 *          the string well-formed UTF-8.
 *          Note: The function is not able to restore corrupted codepoints because
 *          UTF-8 does not contain any information applicable for an error correction.
 * \param  pcs  Referenced cstr.
 * \param  rep  Nullterminated string of characters which is used to replace the
 *              invalid code units. The user is responsible that rep contains
 *              well-formed UTF-8.
 *              If rep is a zero-length string invalid code units will be removed.
 *              If rep is NULL invalid code units will be replaced with the Unicode
 *              replacement character (U+FFFD).
 * \return  The updated cstr. */
cstr cs_utf8_replace_invalids(cstr *pcs, const char *rep);

//////////////////////////////////////////////////////////
// vvv  based on wchar_t  vvv

/** \brief  Type to indicate that the pointer is a cstrw.
 *  Besides of that, it is compatible with a pointer to wchar_t,
 *  and it behaves like a nullterminated C-string of wide characters. */
typedef wchar_t *cstrw;

/** \brief  Type to indicate that the pointer is a cstrw
 *  with immutable content. */
typedef const wchar_t *ccstrw;

//////////////////////////////////////////////////////////

/** \brief  Create and initialize a cstrw from a nullterminated string.
 * \param  strw  Nullterminated string.
 * \return  Newly allocated cstrw containing the copied strw. */
cstrw csw_init(const wchar_t *strw);

/** \brief  Create and initialize a cstrw from formatted data.
 * \param  formatw  Format specification. (supported specifiers like for wprintf)
 * \param  ...      Arguments.
 * \return  Newly allocated cstrw containing the formatted data. NULL if the function fails. */
cstrw csw_init_format(const wchar_t *formatw, ...)
#if defined(__GNUC__) || defined(__clang__)
  __attribute__((__nonnull__(1)))
#endif
  ;

/** \brief  Create and initialize a cstrw from a string of wide characters.
 * \param  seqw  String of wide characters.
 * \param  cnt   First number of wide characters of seqw to be used.
 * \return  Newly allocated cstrw containing the copied substring. */
cstrw csw_init_n(const wchar_t *seqw, size_t cnt);

/** \brief  Create and initialize a cstrw from the entire file content.
 * \param  stream  File stream to be read. Must not be `stdin`
 *                 and should have been opened in binary mode.
 * \return  Newly allocated cstrw containing the copied wide characters.
 *          NULL if the function fails. */
cstrw csw_init_rdentirefile(FILE *stream);

/** \brief  Create and initialize a cstrw from wide characters read out of a file.
 * \param  stream  File stream to be read. Can be `stdin` if cnt is not -1
 *                 when redirected from keyboard.
 * \param  cnt     First number of wide characters of stream to be used.
 *                 Pass -1 to read until EOF.
 *                 Note that cnt begins at the current position of the file cursor.
 * \return  Newly allocated cstrw containing the copied wide characters.
 *          NULL if the function fails to determine the file size. */
cstrw csw_init_rdfile(FILE *stream, size_t cnt);

/** \brief  Create and initialize a cstrw from a line of wide characters read out of a file.
 * \param  stream  File stream to be read until the first occurrence of line feed or EOF.
 *                 Can be `stdin`.
 * \return  Newly allocated cstrw containing the copied wide characters without the line feed
 *          at the end of the line. */
cstrw csw_init_rdline(FILE *stream);

/** \brief  Create a zero-length cstrw with at least 'size' + 1 bytes storage capacity.
 * \param  size  Expected length of the string to be stored.
 * \return  Newly allocated zero-length cstrw. */
cstrw csw_init_zero(size_t size);

/** \brief  Delete the cstrw and release memory back to the operating system.
 * \param  csw  The cstrw to be destroyed. */
void csw_free(ccstrw csw);

/** \brief  Get the maximum number of wide chars that technically can be saved into a cstrw.
 * \return  Maximum number of wide chars, not including the terminating null. */
size_t csw_max_size(void);

//////////////////////////////////////////////////////////

/** \brief  Add data from a nullterminated string of wide characters at the end of a cstrw.
 * \param  pcsw  Referenced cstrw where new data is appended.
 * \param  strw  Nullterminated string of wide characters.
 * \return  The cstrw containing the updated data. */
cstrw csw_append(cstrw *pcsw, const wchar_t *strw);

/** \brief  Add data from a string of wide characters at the end of a cstrw.
 * \param  pcsw  Referenced cstrw where new data is appended.
 * \param  seqw  String of wide characters.
 * \param  cnt   First number of wide characters of seqw to be appended.
 * \return  The cstrw containing the updated data. */
cstrw csw_append_n(cstrw *pcsw, const wchar_t *seqw, size_t cnt);

/** \brief  Assign data from a nullterminated string of wide characters to the cstrw,
 *          replace its current contents.
 * \param  pcsw  Referenced cstrw where the data is assigned.
 * \param  strw  Nullterminated string of wide characters.
 * \return  The cstrw containing the new data. */
cstrw csw_assign(cstrw *pcsw, const wchar_t *strw);

/** \brief  Assign data from a string of wide characters to the cstrw,
*           replace its current contents.
 * \param  pcsw  Referenced cstrw where the data is assigned.
 * \param  seqw  String of wide characters.
 * \param  cnt   First number of characters of seq to be used.
 * \return  The cstrw containing the new data. */
cstrw csw_assign_n(cstrw *pcsw, const wchar_t *seqw, size_t cnt);

/** \brief  Get a wide character in a cstrw from the specified position.
 * \param  csw  The cstrw.
 * \param  pos  Zero-based position of a wide character within the cstrw, clamped to its length.
 * \return  The wide character at the specified position, or L'\0' if out of range. */
wchar_t csw_at(ccstrw csw, size_t pos);

/** \brief  Get the capacity of the allocated wide character buffer - 1 which is reserved for the terminating null.
 * \param  csw  The cstrw.
 * \return  The size of the storage capacity allocated for the cstrw. */
size_t csw_capacity(ccstrw csw);

/** \brief  Erase the contents of the cstrw, which becomes a zero-length string.
 * \param  csw  The cstrw.
 * \return  The updated cstrw. */
cstrw csw_clear(cstrw csw);

/** \brief  Test if the cstrw has zero length.
 * \param  csw  The cstrw.
 * \return  true if the string length is 0, false otherwise. */
bool csw_empty(ccstrw csw);

/** \brief  Erase a part of the cstrw, reducing its length.
 * \param  csw     The cstrw.
 * \param  pos     Zero-based position in the cstrw where the erased portion begins.
 *                 If pos is out of range, the cstrw is unchanged.
 * \param  length  Length of the portion to be erased.
 *                 If the specified span ends out of range, only the valid portion is erased.
 * \return  The updated cstrw. */
cstrw csw_erase(cstrw csw, size_t pos, size_t length);

/** \brief  Update the cstrw to a fixed length by either padding or shortening.
 * \param  pcsw    Referenced cstrw.
 * \param  length  New length of the cstrw.
 *                 If length <= the current length of the cstrw, `pad` is ignored and might be NULL.
 * \param  padw    Nullterminated string containing the wide characters to be padded.
 * \param  mode    Flags which specify where the cstrw is to be padded or shortened.
 *                   1 at the begin of the cstrw
 *                   2 at the end of the cstrw
 *                 Their combination (1|2) leads to a centered alignment.
 * \return  The cstrw containing the updated data. */
cstrw csw_fix(cstrw *pcsw, size_t length, const wchar_t *padw, unsigned mode);

/** \brief  Insert additional wide characters into the cstrw right before the wide character indicated by pos.
 * \param  pcsw  Referenced cstrw where the data is inserted.
 * \param  pos   Zero-based position in the cstrw where the wide characters are inserted.
 *               If pos is out of range, the wide characters are appended.
 * \param  strw  Nullterminated string containing the wide characters to be inserted.
 * \return  The cstrw containing the updated data. */
cstrw csw_insert(cstrw *pcsw, size_t pos, const wchar_t *strw);

/** \brief  Insert additional wide characters into the cstrw right before the wide character indicated by pos.
 * \param  pcsw  Referenced cstrw where the data is inserted.
 * \param  pos   Zero-based position in the cstrw where the wide characters are inserted.
 *               If pos is out of range, the wide characters are appended.
 * \param  seqw  String of wide characters.
 * \param  cnt   First number of wide characters of seqw to be inserted.
 * \return  The cstrw containing the updated data. */
cstrw csw_insert_n(cstrw *pcsw, size_t pos, const wchar_t *seqw, size_t cnt);

/** \brief  Get the length of the cstrw.
 * \param  csw  The cstrw.
 * \return  The length of the cstrw, without the terminating null-character. */
size_t csw_length(ccstrw csw);

/** \brief  Erases the last character of the cstrw.
 * \param  csw  The cstrw.
 * \return  The wide character removed from the string end, or L'\0' if the cstrw has zero length. */
wchar_t csw_pop_back(cstrw csw);

/** \brief  Append a wide character to the end of a cstrw.
 * \param  pcsw  Referenced cstrw where the wide character is appended.
 * \param  chw   Wide character to be appended.
 * \return  The cstrw containing the updated data. */
cstrw csw_push_back(cstrw *pcsw, const wchar_t chw);

/** \brief  Replace the portion of the cstrw that begins at wide character 'pos'
 *          and spans 'length' wide characters.
 * \param  pcsw    Referenced cstrw where data is replaced.
 * \param  pos     Zero-based position in the cstrw where the portion begins.
 *                 If pos is out of range, the replacement string appended.
 * \param  length  Length of the portion to be replaced.
 *                 If the portion ends out of range, the cstr will be extended.
 * \param  strw    Nullterminated string containing the wide characters the portion is replaced with.
 * \return  The cstrw containing the updated data. */
cstrw csw_replace(cstrw *pcsw, size_t pos, size_t length, const wchar_t *strw);

/** \brief  Replace the portion of the cstrw that begins at wide character 'pos' and spans 'length' wide characters.
 * \param  pcsw    Referenced cstrw where data is replaced.
 * \param  pos     Zero-based position in the cstrw where the portion begins.
 *                 If pos is out of range, the replacement string appended.
 * \param  length  Length of the portion to be replaced.
 *                 If the portion ends out of range, the cstrw will be extended.
 * \param  seqw    String containing the wide characters the portion is replaced with.
 * \param  cnt     First number of wide characters of seqw to be used.
 * \return  The cstrw containing the updated data. */
cstrw csw_replace_n(cstrw *pcsw, size_t pos, size_t length, const wchar_t *seqw, size_t cnt);

/** \brief  Allocate at least the requested capacity.
 * \param  pcsw  Referenced cstrw.
 * \param  capa  Minimum capacity of the cstrw.
 * \return  The updated cstrw. */
cstrw csw_reserve(cstrw *pcsw, size_t capa);

/** \brief  Resize the cstrw to a length of 'size' wide characters.
 * \param  pcsw   Referenced cstrw to be resized.
 * \param  size   New size of the cstrw.
 * \param  fillw  Character used to fill the new wide character space added to the cstrw.
 *                If the new size is smaller than the current size, fillw is ignored.
 * \return  The updated cstrw. */
cstrw csw_resize(cstrw *pcsw, size_t size, wchar_t fillw);

/** \brief  Reverse the wide character order in the specified cstrw.
 * \param  csw  The cstrw.
 * \return  The updated cstrw. */
cstrw csw_reverse(cstrw csw);

/** \brief  Requests the cstrw to reduce its capacity to fit its size.
 * \param  pcsw  Referenced cstrw.
 * \return  The updated cstrw. */
cstrw csw_shrink_to_fit(cstrw *pcsw);

/** \brief  Get the length of the cstrw.
 * \param  csw  The cstrw.
 * \return  The length of the cstrw, without the terminating null-character. */
size_t csw_size(ccstrw csw);

/** \brief  Newly construct a cstrw with its value initialized to a copy of a substring of 'csw'.
 * \param  csw     The cstrw.
 * \param  pos     Zero-based position in the cstrw where the substring begins.
 *                 If pos is out of range, a zero-length cstrw is returned.
 * \param  length  Length of the portion to be copied.
 *                 If the specified span ends out of range, the substring contains only the valid portion.
 * \return  The new cstrw containing the copied substring. */
cstrw csw_substr(ccstrw csw, size_t pos, size_t length);

/** \brief  Remove specified prepended and/or appended wide characters from a cstrw.
 * \param  csw        The cstrw.
 * \param  char_setw  Nullterminated string of wide characters which are to be removed.
 * \param  mode       Flags which specify where the wide characters are to be removed.
 *                      1 to remove prepended wide characters
 *                      2 to remove appended wide characters
 * \return  The updated cstrw. */
cstrw csw_trim(cstrw csw, const wchar_t *char_setw, unsigned mode);

/** \brief  Force the cstrw to update its size property to the specified length.
 *          Alteration of a cstrw outside of the library is supported as long as this does neither violate
 *          the boundaries of the allocated storage capacity (see csw_capacity), nor its memory address.
 *          In those cases it's necessary to adjust the internal size property of the cstrw accordingly.
 * \param  csw     The cstrw.
 * \param  length  Actual length of the cstrw.
 * \return  The updated cstrw with a terminating null set. */
cstrw csw_update_length(cstrw csw, size_t length);

#endif // CSTR_H_INCLUDED__
