/*
Copyright (c) 2020 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef CVEC_H_INCLUDED__
#define CVEC_H_INCLUDED__

#include <stdbool.h>
#include <stddef.h>

//////////////////////////////////////////////////////////

// Allocated pointers point to mutable memory by design, regardless of any const declaration in the code. A simple cast to `void*` may still cause a warning though.
static inline void *discard_const_impl_detail(const void *ptr)
{
  const union
  {
    const void *ptr2const;
    void *ptr2mutable;
  } discardconst = { ptr };

  return discardconst.ptr2mutable;
}

/** \brief  Macro to cast a pointer to void (cvec or ccvec) to a pointer to the type specified by _elem_type. */
#define PTRTO(_cvec, _elem_type) ((_elem_type *)discard_const_impl_detail(_cvec))

/** \brief  Macro to create a pointer to _elem_type, decayed from an array literal out of the values passed to the ellipsis. */
#define MAKEPTR(_elem_type, ...) (&*(const _elem_type[]){ __VA_ARGS__ })

//////////////////////////////////////////////////////////

/** \brief  Type to indicate that the pointer is a cvec. */
typedef void *cvec;

/** \brief  Type to indicate that the pointer is a cvec with immutable content. */
typedef void *ccvec;

/** \brief  Type of a callback function passed to cv_for_each. */
typedef bool (*cv_for_each_proc_t)(void *element_data, void *user_param);
// Declaration of a callback function:
/*
  bool for_each_callback(void *p_data, void *user_param);
*/
// for_each_callback is a placeholder for the application-defined function name.
// The first parameter is the pointer to the data saved in the cvec element.
// The second parameter is the pointer passed to the last parameter of cv_for_each().
// Return true to continue the enumeration or false otherwise.
// +++ NOTE +++ The behavior is undefined if the callback function alters the number of elements or their order in the cvec.

//////////////////////////////////////////////////////////

/** \brief  Create and initialize a cvec from an array.
 * \param  seq          Array.
 * \param  cnt          First number of elements of seq to be used.
 * \param  elementsize  Size of the element type in the array.
 * \return  Newly allocated cvec containing the copied elements. */
cvec cv_init(const void *seq, size_t cnt, size_t elementsize);

/** \brief  Create a zero-length cvec with at least 'size' elements of storage capacity.
 * \param  size  Expected size of the vector to be stored.
 * \param  elementsize  Size of the element type in the vector.
 * \return  Newly allocated zero-length vector. */
cvec cv_init_zero(size_t size, size_t elementsize);

/** \brief  Delete the cvec and release memory back to the operating system.
 * \param  cv  The cvec to be destroyed. */
void cv_free(ccvec cv);

/** \brief  Get the maximum number of elements that technically can be saved into a cvec.
 * \param  elementsize  Size of the element type in the vector.
 * \return  Maximum number of elements, not including the terminating null. */
size_t cv_max_size(size_t elementsize);

//////////////////////////////////////////////////////////

/** \brief  Add data from an array at the end of a cvec.
 * \param  pcv  Referenced cvec where new data is appended.
 * \param  seq  Array.
 * \param  cnt  First number of elements of seq to be appended.
 * \return  The cvec containing the updated data. */
cvec cv_append(cvec *pcv, const void *seq, size_t cnt);

/** \brief  Assign data from am array to the cvec, replace its current contents.
 * \param  pcv  Referenced cvec where the data is assigned.
 * \param  seq  Array.
 * \param  cnt  First number of elements of seq to be used.
 * \return  The cvec containing the new data. */
cvec cv_assign(cvec *pcv, const void *seq, size_t cnt);

/** \brief  Get a pointer to an element in a cvec from the specified position.
 * \param  cv   The cvec.
 * \param  pos  Zero-based position of an element within the cvec, clamped to its length.
 * \return  The pointer to the element at the specified position, or NULL if out of range. */
const void *cv_at(ccvec cv, size_t pos);

/** \brief  Get the capacity of the allocated buffer.
 * \param  cv  The cvec.
 * \return  The size of the storage capacity allocated for the cvec as number of elements. */
size_t cv_capacity(ccvec cv);

/** \brief  Erase the contents of the cvec, which becomes a zero-length vector.
 * \param  cv  The cvec.
 * \return  The updated cvec. */
cvec cv_clear(cvec cv);

/** \brief  Test if the cvec has zero length.
 * \param  cv  The cvec.
 * \return  true if the vector length is 0, false otherwise. */
bool cv_empty(ccvec cv);

/** \brief  Erase a part of the cvec, reducing its size.
 * \param  cv   The cvec.
 * \param  pos  Zero-based position in the cvec where the erased portion begins.
 *              If pos is out of range, the cvec is unchanged.
 * \param  cnt  Number of elements of the portion to be erased.
 *              If the specified span ends out of range, only the valid portion is erased.
 * \return  The updated cvec. */
cvec cv_erase(cvec cv, size_t pos, size_t cnt);

/** \brief  Function to enumerate the values of a range of elements.
 * \param  cv             The cvec.
 * \param  pos            Zero-based position in the cvec where the enumeration begins.
 * \param  cnt            Number of elements of the portion to be enumerated.
 *                        If the specified span ends out of range, only the valid portion is enumerated.
 *                        Pass -1 to enumerate all remaining elements.
 * \param  callback_func  Application-defined function which is called for each element in the range
 *                        defined by pos and count.
 * \param  user_param     Pointer to user-defined data which gets passed to the second parameter of
 *                        callback_func.
 * \return  `true` if the enumeration was executed, `false` if pos is out of range or if cnt is zero. */
bool cv_for_each(cvec cv, size_t pos, size_t cnt, cv_for_each_proc_t callback_func, void *user_param);

/** \brief  Insert additional elements into the cvec right before the element indicated by pos.
 * \param  pcv  Referenced cvec where the data is inserted.
 * \param  pos  Zero-based position in the cvec where the elements are inserted.
 *              If pos is out of range, the elements are appended.
 * \param  seq  Array containing the elements to be inserted.
 * \param  cnt  First number of elements of seq to be inserted.
 * \return  The cvec containing the updated data. */
cvec cv_insert(cvec *pcv, size_t pos, const void *seq, size_t cnt);

/** \brief  Erases the last element of the cvec.
 * \param  cv  The cvec.
 * \return  Pointer to the element removed from the vector, or NULL if the cvec has zero length. */
const void *cv_pop_back(cvec cv);

/** \brief  Append an element to the end of a cvec.
 * \param  pcv   Referenced cvec where the element is appended.
 * \param  elem  Pointer to the element to be appended.
 * \return  The cvec containing the updated data. */
cvec cv_push_back(cvec *pcv, const void *elem);

/** \brief  Allocate at least the requested capacity.
 * \param  pcv   Referenced cvec.
 * \param  capa  Minimum capacity of the cvec as number of elements.
 * \return  The updated cvec. */
cvec cv_reserve(cvec *pcv, size_t capa);

/** \brief  Resize the cvec to a length of 'size' elements.
 * \param  pcv   Referenced cvec to be resized.
 * \param  size  New size of the cvec.
 * \param  elem  Pointer to the element used to fill the new space added to the cvec.
 *               If the new size is smaller than the current size, elem is ignored.
 *               If NULL is passed, the space added is zero-initialized.
 * \return  The updated cvec. */
cvec cv_resize(cvec *pcv, size_t size, const void *elem);

/** \brief  Request the cvec to reduce its capacity to fit its size.
 * \param  pcv  Referenced cvec to be resized.
 * \return  The updated cvec. */
cvec cv_shrink_to_fit(cvec *pcv);

/** \brief  Get the size of the cvec.
 * \param  cv  The cvec.
 * \return  The size of the cvec. */
size_t cv_size(ccvec cv);

/** \brief  Force the cvec to update its size property to the specified size.
 *          Alteration of a cvec outside of the library is supported as long as this does neither violate
 *          the boundaries of the allocated storage capacity (see cv_capacity), nor its memory address.
 *          In those cases it's necessary to adjust the internal size property of the cvec accordingly.
 * \param  cv    The cvec.
 * \param  size  Actual size of the cvec.
 * \return  The updated cvec. */
cvec cv_update_size(cvec cv, size_t size);

#endif // CVEC_H_INCLUDED__
