/*
Copyright (c) 2020 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "memmem.h"

#if defined(_MSC_VER) && _MSC_VER >= 1914
#  pragma warning(push)
#  pragma warning(disable : 5045) // Information that MS compiler will insert Spectre mitigation if /Qspectre switch specified
#endif

// Allocated pointers point to mutable memory by design, regardless of any const declaration in the code.
// A simple cast to `void*` may still cause a warning though. Besides of that (in tradition of standard
// functions like memchr, strchr, strpbrk, strrchr, strstr) the return value of the memmem function is a
// pointer to mutable even if the first parameter is a pointer to const. This is basically wrong but the
// behavior of memmem and its siblings shall be aligned with those functions.
static inline void *cast_to_nonconst(const void *ptr)
{
  const union
  {
    const void *ptr2const;
    void *ptr2mutable;
  } discardconst = { ptr };

  return discardconst.ptr2mutable;
}

void *memmem(const void *buffer, size_t buffer_size, const void *target, size_t target_size)
{
  if (!buffer || !target || target_size > buffer_size || !buffer_size || !target_size)
    return NULL;

  if (target_size == buffer_size)
  {
    if (memcmp(buffer, target, target_size) == 0)
      return cast_to_nonconst(buffer);

    return NULL;
  }

  const unsigned char *buffer_cursor = (const unsigned char *)buffer, *target_cursor = (const unsigned char *)target;
  size_t buffer_hash = 0U, target_hash = 0U, factor = 1U;
  for (const unsigned char *end = target_cursor + target_size - 1; target_cursor < end; ++buffer_cursor, ++target_cursor)
  {
    buffer_hash = ((buffer_hash << 1) + *buffer_cursor);
    target_hash = ((target_hash << 1) + *target_cursor);
    factor <<= 1;
  }

  buffer_hash = ((buffer_hash << 1) + *buffer_cursor);
  target_hash = ((target_hash << 1) + *target_cursor);

  buffer_cursor = (const unsigned char *)buffer;
  for (const unsigned char *stop = buffer_cursor + buffer_size - target_size; buffer_cursor < stop; ++buffer_cursor)
  {
    if (target_hash == buffer_hash && memcmp(buffer_cursor, target, target_size) == 0)
      return cast_to_nonconst(buffer_cursor);

    buffer_hash = ((buffer_hash - *buffer_cursor * factor) << 1) + *(buffer_cursor + target_size);
  }

  if (target_hash == buffer_hash && memcmp(target, buffer_cursor, target_size) == 0)
    return cast_to_nonconst(buffer_cursor);

  return NULL;
}

void *memrmem(const void *buffer, size_t buffer_size, const void *target, size_t target_size)
{
  if (!buffer || !target || target_size > buffer_size || !buffer_size || !target_size)
    return NULL;

  if (target_size == buffer_size)
  {
    if (memcmp(buffer, target, target_size) == 0)
      return cast_to_nonconst(buffer);

    return NULL;
  }

  const unsigned char *buffer_cursor = (const unsigned char *)buffer + buffer_size - 1, *target_cursor = (const unsigned char *)target + target_size - 1;
  size_t buffer_hash = 0U, target_hash = 0U, factor = 1U;
  for (; target_cursor > (const unsigned char *)target; --buffer_cursor, --target_cursor)
  {
    buffer_hash = ((buffer_hash << 1) + *buffer_cursor);
    target_hash = ((target_hash << 1) + *target_cursor);
    factor <<= 1;
  }

  buffer_hash = ((buffer_hash << 1) + *buffer_cursor);
  target_hash = ((target_hash << 1) + *target_cursor);

  for (; buffer_cursor > (const unsigned char *)buffer; --buffer_cursor)
  {
    if (target_hash == buffer_hash && memcmp(buffer_cursor, target, target_size) == 0)
      return cast_to_nonconst(buffer_cursor);

    buffer_hash = ((buffer_hash - *(buffer_cursor + target_size - 1) * factor) << 1) + *(buffer_cursor - 1);
  }

  if (target_hash == buffer_hash && memcmp(target, buffer_cursor, target_size) == 0)
    return cast_to_nonconst(buffer_cursor);

  return NULL;
}

char *strrstr(const char *str, const char *substr)
{
  if (!str || !substr)
    return NULL;

  return (char *)memrmem((const void *)str, strlen(str), (const void *)substr, strlen(substr));
}

wchar_t *wcsrstr(const wchar_t *strw, const wchar_t *substrw)
{
  if (!strw || !substrw)
    return NULL;

  const size_t strw_size = wcslen(strw), substrw_size = wcslen(substrw);
  if (substrw_size > strw_size || !strw_size || !substrw_size)
    return NULL;

  if (substrw_size == strw_size)
  {
    if (memcmp(strw, substrw, substrw_size * sizeof(wchar_t)) == 0)
      return cast_to_nonconst(strw);

    return NULL;
  }

  const wchar_t *strw_cursor = strw + strw_size - 1, *substrw_cursor = substrw + substrw_size - 1;
  size_t strw_hash = 0U, substrw_hash = 0U, factor = 1U;
  for (; substrw_cursor > substrw; --strw_cursor, --substrw_cursor)
  {
    strw_hash = (strw_hash << 1) + (unsigned)(*strw_cursor);
    substrw_hash = (substrw_hash << 1) + (unsigned)(*substrw_cursor);
    factor <<= 1;
  }

  strw_hash = (strw_hash << 1) + (unsigned)(*strw_cursor);
  substrw_hash = (substrw_hash << 1) + (unsigned)(*substrw_cursor);

  for (; strw_cursor > strw; --strw_cursor)
  {
    if (substrw_hash == strw_hash && memcmp(strw_cursor, substrw, substrw_size * sizeof(wchar_t)) == 0)
      return cast_to_nonconst(strw_cursor);

    strw_hash = ((strw_hash - (unsigned)(*(strw_cursor + substrw_size - 1)) * factor) << 1) + (unsigned)(*(strw_cursor - 1));
  }

  if (substrw_hash == strw_hash && memcmp(substrw, strw_cursor, substrw_size * sizeof(wchar_t)) == 0)
    return cast_to_nonconst(strw_cursor);

  return NULL;
}

#if defined(_MSC_VER) && _MSC_VER >= 1914
#  pragma warning(pop) // C5045 Spectre mitigation
#endif
